/* droid-sans-400normal - latin */
@font-face {
  font-family: 'Droid Sans';
  font-style: normal;
  font-display: swap;
  font-weight: 400;
  src:
    local('Droid Sans Regular '),
    local('Droid Sans-Regular'),
    url('./files/droid-sans-latin-400.woff2') format('woff2'), /* Super Modern Browsers */
    url('./files/droid-sans-latin-400.woff') format('woff'); /* Modern Browsers */
}

/* droid-sans-700normal - latin */
@font-face {
  font-family: 'Droid Sans';
  font-style: normal;
  font-display: swap;
  font-weight: 700;
  src:
    local('Droid Sans Bold '),
    local('Droid Sans-Bold'),
    url('./files/droid-sans-latin-700.woff2') format('woff2'), /* Super Modern Browsers */
    url('./files/droid-sans-latin-700.woff') format('woff'); /* Modern Browsers */
}

